from typing import Any, Callable, TypeVar

from .core import CurrentGraph, RegistrationFilter, Resolver, default_registration_filter


def use_registered(cls: type, filter: RegistrationFilter = default_registration_filter):
    def factory(resolver: Resolver):
        return resolver.resolve(cls, filter=filter)

    return factory


def use_registered_async(cls: type, filter: RegistrationFilter = default_registration_filter):
    async def factory(resolver: Resolver):
        return await resolver.resolve_async(cls, filter=filter)

    return factory


def use_from_current_graph(cls: type, filter: RegistrationFilter = default_registration_filter):
    def factory(current_graph: CurrentGraph):
        return current_graph.resolve(cls, filter=filter)

    return factory


def use_from_current_graph_async(cls: type, filter: RegistrationFilter = default_registration_filter):
    async def factory(current_graph: CurrentGraph):
        return await current_graph.resolve_async(cls, filter=filter)

    return factory


T = TypeVar("T")


def create_type_mapping(
    service_type: type[T],
    key_getter: Callable[[T], Any],
    filter: RegistrationFilter = default_registration_filter,
):
    def factory(resolver: Resolver):
        items = resolver.resolve(list[service_type], filter=filter)
        return {key_getter(item): item for item in items}

    return factory


def create_type_mapping_async(
    service_type: type[T],
    key_getter: Callable[[T], Any],
    filter: RegistrationFilter = default_registration_filter,
):
    async def factory(resolver: Resolver):
        items = await resolver.resolve_async(list[service_type], filter=filter)
        return {key_getter(item): item for item in items}

    return factory



import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import ClassVar
from uuid import uuid4

from clean_ioc import Container

logger = logging.getLogger(__name__)


class BaseBundle(ABC):
    @abstractmethod
    def apply(self, container: Container): ...

    def __call__(self, container: Container):
        self.apply(container=container)


class RunOnceBundle(BaseBundle):
    BUNDLE_RUN_HISTORY: ClassVar[dict[str, list[str]]] = defaultdict(list)

    @abstractmethod
    def apply(self, container: Container): ...

    @abstractmethod
    def get_bundle_identifier(self) -> str: ...

    def __call__(self, container: Container):
        bundle_identifier = self.get_bundle_identifier()
        bundle_containers = self.__class__.BUNDLE_RUN_HISTORY[bundle_identifier]
        container_id = container.id

        if container_id in bundle_containers:
            logger.debug("Bundle %s attempted to run more than once on container %s", bundle_identifier, container_id)
            return

        self.apply(container=container)
        bundle_containers.append(container_id)


class OnlyRunOncePerInstanceBundle(RunOnceBundle):
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._instance_id = str(uuid4())  # type: ignore
        return instance

    def get_bundle_identifier(self) -> str:
        module = self.__class__.__module__
        class_name = self.__class__.__name__
        return f"{module}.{class_name}-{self._instance_id}"  # type: ignore


class OnlyRunOncePerClassBundle(RunOnceBundle):
    def get_bundle_identifier(self) -> str:
        module = self.__class__.__module__
        class_name = self.__class__.__name__
        return f"{module}.{class_name}"



import inspect
from collections.abc import Iterable
from typing import Callable, TypeVar

from theutilitybelt.functional.predicate import predicate
from theutilitybelt.functional.utils import constant

from .core import Lifespan, Registration

all_registrations = constant(True)


def create_filter(func: Callable[[Registration], bool]):
    return predicate(func)


def with_name(name: str | None):
    """
    Filter registrations equal the name
    """

    def _with_name(r: Registration):
        return r.name == name

    return predicate(_with_name)


def name_starts_with(prefix: str):
    """
    Filter registrations where the name starts the prefix
    """

    def _name_starts_with(r: Registration):
        if r.name is not None:
            return r.name.startswith(prefix)
        return False

    return predicate(_name_starts_with)


def name_ends_with(suffix: str):
    """
    Filter registrations where the name ends the suffix
    """

    def _name_ends_with(r: Registration):
        if r.name is not None:
            return r.name.endswith(suffix)
        return False

    return predicate(_name_ends_with)


is_not_named = with_name(None)
is_not_named.__doc__ = "Filter for registrations that do not have a name"

is_named = ~is_not_named
is_named.__doc__ = "Filter for registrations that have a name"


def with_implementation(implementation: type):
    """
    Filter to registrations that have the implementation
    """

    def _with_implementation(r: Registration):
        return r.implementation == implementation

    return predicate(_with_implementation)


def with_implementation_matching_filter(type_filter: Callable[[type], bool]):
    """
    Returns a filter function that checks if the implementation of a given registration matches a specified type filter.

    Parameters:
        type_filter (Callable[[type], bool]): A function that takes a type and returns a boolean indicating whether
        the type matches a specific criteria.

    Returns:
        Callable[[Registration], bool]: A filter function that takes a registration and returns True if
        the implementation of the registration matches the specified type filter, False otherwise.

    Example:
        >>> type_filter = lambda x: issubclass(x, int)
        >>> registration = Registration(implementation=123)
        >>> with_implementation_matching_filter(type_filter)(registration)
        True
    """

    def _with_implementation_matching_filter(r: Registration):
        if inspect.isfunction(r.implementation):
            return False

        return type_filter(r.implementation)  # type: ignore

    return predicate(_with_implementation_matching_filter)


def has_generic_args_matching(pair: tuple[TypeVar | str, type]):
    """
    Filter registrations where generic type arg matches
    >>> TBar = TypeVar("TBar")
    >>> class Foo(Generic[TBar]):
    ...    pass
    >>> class IntFoo(Foo[int]):
    ...    pass
    >>> class StringFoo(Foo[str]):
    ...    pass
    >>> container.register(Foo, IntFoo)
    >>> container.register(Foo, StringFoo)
    >>> container.resolve(Foo, filter=has_generic_args_matching((TBar, int))) # returns instance of IntFoo
    """

    def _has_generic_args_matching(r: Registration):
        return r.generic_mapping.get(pair[0]) == pair[1]

    return predicate(_has_generic_args_matching)


def has_tag(name: str, value: str | None = None):
    """
    Check if a given registration has a specific tag.

    Parameters:
        name (str): The name of the tag to check for.
        value (str | None, optional): The value of the tag to check for. Defaults to None.

    Returns:
        Callable[[Registration], bool]: A filter function that takes a registration and returns True
        if the registration has the specified tag, False otherwise.
    """

    def _has_tag(r: Registration):
        return r.has_tag(name, value)

    return predicate(_has_tag)


def has_tag_with_value_or_missing_tag(name: str, value: str):
    """
    Check if a given registration has a specific tag with a given value or if it does not have that tag.

    Parameters:
        name (str): The name of the tag to check for.
        value (str): The value of the tag to check for.

    Returns:
        Callable[[Registration], bool]: A filter function that takes a registration and returns True
        if the registration has the specified tag with the given value or if it does not have that tag, False otherwise.
    """
    return has_tag(name, value) | ~has_tag(name)


def has_tag_with_value_in(name: str, *values: str):
    """
    Check if a given registration has a specific tag with a value that matches any of the given values.

    Parameters:
        name (str): The name of the tag to check for.
        *values (str): The values to check for.

    Returns:
        Callable[[Registration], bool]: A filter function that takes a registration and returns True if
        the registration has the specified tag with a value that matches any of the given values, False otherwise.
    """
    predicates = [has_tag(name, v) for v in values]

    def _has_tag_with_value_in(r: Registration):
        return any([p(r) for p in predicates])

    return predicate(_has_tag_with_value_in)


def has_lifespan(lifespan: Lifespan):
    """
    Check if a given registration has a specific lifespan.

    Parameters:
        lifespan (str): The lifespan to check for.

    Returns:
        Callable[[Registration], bool]: A filter function that takes a registration and returns True
        if the registration has the specified lifespan, False otherwise.
    """

    def _has_lifespan(r: Registration):
        return r.lifespan == lifespan

    return predicate(_has_lifespan)


def has_lifespan_in(lifespans: Iterable[Lifespan]):
    """
    Returns a predicate function that checks if the lifespan of a registration is in the given lifespans.

    Args:
        lifespans (Iterable[Lifespan]): The lifespans to check against.

    Returns:
        Callable[[Registration], bool]: A predicate function that returns True if the lifespan of a registration
        is in the given lifespans, False otherwise.
    """

    def _has_lifespans(r: Registration):
        return r.lifespan in lifespans

    return predicate(_has_lifespans)



from collections.abc import Callable
from typing import TypeVar

from theutilitybelt.functional.predicate import predicate

TArg = TypeVar("TArg")


def constant(val: TArg) -> Callable[..., TArg]:
    def fun(*_):
        return val

    return fun


def fn_not(func: Callable[..., bool]):
    def predicate(*args, **kwargs):
        return not func(*args, **kwargs)

    return predicate


def fn_and(*funcs: Callable[..., bool]):
    def predicate(*args, **kwargs):
        for func in funcs:
            if not func(*args, **kwargs):
                return False
        return True

    return predicate


def fn_or(*funcs: Callable[..., bool]):
    def predicate(*args, **kwargs):
        for func in funcs:
            if func(*args, **kwargs):
                return True
        return False

    return predicate


def fn_xor(*funcs: Callable[..., bool]):
    def predicate(*args, **kwargs):
        return len([f for f in funcs if f(*args, **kwargs)]) % 2 == 1

    return predicate


Predicate = predicate



from typing import Callable

from theutilitybelt.functional.predicate import predicate
from theutilitybelt.functional.utils import constant

from .core import Node, NodeFilter

yes = constant(True)


def implementation_type_is(cls: type):
    def inner(node: Node):
        return node.implementation == cls

    return predicate(inner)


def service_type_matches_type_filter(type_filter: Callable[[type], bool]):
    def inner(node: Node):
        return type_filter(node.service_type)

    return predicate(inner)


def implementation_matches_type_filter(type_filter: Callable[[type], bool]):
    def inner(node: Node):
        return type_filter(node.implementation_type)  # type: ignore

    return predicate(inner)


def service_type_is(cls: type):
    def inner(node: Node):
        return node.service_type == cls

    return predicate(inner)


def registration_name_is(name: str):
    def inner(node: Node):
        return node.registration_name == name

    return predicate(inner)


def has_registration_tag(name: str, value: str | None = None):
    def inner(node: Node):
        return node.has_registration_tag(name, value)

    return predicate(inner)


def has_dependant_service_type(service_type: type):
    def inner(node: Node):
        return node.has_dependant_service_type(service_type)

    return predicate(inner)


def has_dependant_implementation_type(implementation_type: type):
    def inner(node: Node):
        return node.has_dependant_implementation_type(implementation_type)

    return predicate(inner)


def has_dependant_instance_type(instance_type: type):
    def inner(node: Node):
        return node.has_dependant_instance_type(instance_type)

    return predicate(inner)


def jump_parent(filter: NodeFilter):
    def inner(node: Node):
        return filter(node.parent)

    return predicate(inner)



import inspect

from theutilitybelt.functional.predicate import predicate


def named(name: str):
    """
    A function that takes a string 'name' and returns a predicate function that checks if the input type's
    name matches the given 'name'.
    """

    def _named(t: type):
        return t.__name__ == name

    return predicate(_named)


def name_starts_with(name: str):
    def inner(t: type):
        return t.__name__.startswith(name)

    return predicate(inner)


def name_end_with(name: str):
    def inner(t: type):
        return t.__name__.endswith(name)

    return predicate(inner)


def is_in_module(*module_names: str):
    def inner(t: type):
        return any([t.__module__ == m for m in module_names])

    return predicate(inner)


def _is_abstract(t: type):
    return inspect.isabstract(t)


is_abstract = predicate(_is_abstract)


def is_subclass_of(cls: type):
    def inner(t: type):
        return issubclass(t, cls)

    return predicate(inner)
